{% extends "layout.html" %}
{% block title %}{{ _('Password reset request') }}{% endblock %}
{% block preheader %}{{ _('Use this link to reset your password. The link is only valid for %(expiration_delay)s.', expiration_delay=expiration_delay) }}{% endblock %}
{% block content %}<p>{{ _('You recently requested to reset your password for your FitTrackee account.') }} {{ _('Use the button below to reset it.') }}
                          <strong>{{ _('This password reset link is only valid for %(expiration_delay)s.', expiration_delay=expiration_delay) }}</strong>
                        </p>
                        <table class="body-action" align="center" width="100%" cellpadding="0" cellspacing="0" role="presentation">
                          <tr>
                            <td align="center">
                              <table width="100%" border="0" cellspacing="0" cellpadding="0" role="presentation">
                                <tr>
                                  <td align="center">
                                    <a href="{{password_reset_url}}" class="f-fallback button button--green" target="_blank">{{ _('Reset your password') }}</a>
                                  </td>
                                </tr>
                              </table>
                            </td>
                          </tr>
                        </table>{% endblock %}
{% block not_initiated %}{{ _('If you did not request a password reset, please ignore this email.') }}{% endblock %}
{% block url_to_paste %}<table class="body-sub" role="presentation">
                          <tr>
                            <td>
                              <p class="f-fallback sub">{{ _("If you're having trouble with the button above, copy and paste the URL below into your web browser.") }}</p>
                              <p class="f-fallback sub">{{password_reset_url}}</p>
                            </td>
                          </tr>
                        </table>{% endblock %}